/**
  *  \file game/ship-xfer.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ship.h"
#include "game/spec.h"
#include "game/cargo.h"
#include "util/nls.h"
#include "game/planet.h"

/***************************** GShipTransfer *****************************/

GShipTransfer::GShipTransfer(GShip& ship)
    : ship(ship)
{
    setFlags(can_Unload);
    ship.getCargoArbiter().incUseCount();
    if (!ship.isPlayed())
        throw GError(GError::eNotPlaying, _("You do not own this ship"));
}

GShipTransfer::~GShipTransfer()
{
    ship.getCargoArbiter().decUseCount();
    if (!isFinished())
        doCancel();
}

int32
GShipTransfer::getCargoOnObject(GCargoType type)
{
    return ship.getCargo(type);
}

void
GShipTransfer::changeCargoOnObject(GCargoType type, int32 amount)
{
    ship.changeCargoRaw(type, amount);
}

void
GShipTransfer::changeReservedCargo(GCargoType type, int32 amount)
{
    ship.getCargoArbiter().changeReservedCargo(type, amount);
}

bool
GShipTransfer::canHaveCargo(GCargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Colonists:
     case el_Supplies:
     case el_Money:
        return true;
     case el_Fighters:
        return (ship.getBayCount() > 0);
     default:
        return ship.getTorpLauncherCount() > 0
            && cargo_id == getCargoTypeFromTorpType(ship.getTorpType());
    }
}

int32
GShipTransfer::getMaxCargo(GCargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
        if (hasFlag(can_Overload))
            return 10000;
        else
            return ship.getHull().getFuel();
     case el_Money:
        return 10000;
     default:
        /* this is pretty hairy, but probably the best we can do */
        int32 available = hasFlag(can_Overload) ? 20000 : ship.getHull().getCargo();
        int32 rest = available
            - getCargo(el_Tritanium) - getCargo(el_Duranium)
            - getCargo(el_Molybdenum) - getCargo(el_Supplies)
            - getCargo(el_Colonists);
        if (ship.getTorpLauncherCount() > 0)
            rest -= getCargo(getCargoTypeFromTorpType(ship.getTorpType()));
        if (ship.getBayCount() > 0)
            rest -= getCargo(el_Fighters);
        rest += getCargo(cargo_id);
        if (rest > 10000)
            return 10000;
        else
            return rest;
    }
}

string_t
GShipTransfer::getName() const
{
    return ship.getName();
}

/**************************** Ship Transporter ***************************/

/* FIXME: For "undo jettison ammo", we must even lock this location in
   space, not just "this transporter". */

/* Initializing 'reserve()' ought to generate zero-initialisation
   of 'reserve', but Borland (as usual) does not care. */
static const TShipTransfer null_transfer = { {0,0,0,0}, 0, 0, 0 };

/** Construct a transporter which transports to /other/.
    \param ship       the "source" ship
    \param which_one  type of target transporter
    \param other      "target" object
    \pre type of other matches ship_ship
    \pre cargo transporter is empty or already transferring to /other/
    \pre respective facility of ship not already in use */
GShipTransporter::GShipTransporter(GShip& ship, GShip::Transporter which_one, GObject& other)
    : ship(ship), transfer(ship.getTransporter(which_one)), other(&other),
      which_one(which_one), reserve(null_transfer)
{
    if (which_one == GShip::t_Transfer) {
        setFlags(can_Unload);
    } else {
        setFlags(can_GetUnload);
    }

    ASSERT(which_one == GShip::t_Transfer ? dynamic_cast<GShip*>(this->other) != 0 : dynamic_cast<GPlanet*>(this->other) != 0);
    ASSERT(transfer.id == other.getId() || transfer.id == 0);
    ASSERT(!ship.isTransporterInUse(which_one));

    ship.getCargoArbiter().incUseCount();
    ship.setTransporterInUse(which_one, true);

    transfer.id = other.getId();
}

/** Construct a jettison transporter for /ship/.
    \param ship      ship
    \pre sf_Unload facility not already in use */
GShipTransporter::GShipTransporter(GShip& ship)
    : ship(ship), transfer(ship.getTransporter(GShip::t_Unload)), other(0),
      which_one(GShip::t_Unload), reserve(null_transfer)
{
    ASSERT(transfer.id == 0);
    ASSERT(!ship.isTransporterInUse(which_one));

    ship.getCargoArbiter().incUseCount();
    ship.setTransporterInUse(which_one, true);
}

GShipTransporter::~GShipTransporter()
{
    if (!ship.isTransporterActive(which_one)) // ????? why?
        ship.cancelTranporter(which_one);

    ship.setTransporterInUse(which_one, false);
    ship.getCargoArbiter().decUseCount();

    if (!isFinished())
        doCancel();
}

int32
GShipTransporter::getCargoOnObject(GCargoType type)
{
    if (type <= 3)
        return transfer.ore[type] - reserve.ore[type];
    else if (type == el_Colonists)
        return transfer.colonists - reserve.colonists;
    else if (type == el_Supplies)
        return transfer.supplies - reserve.supplies;
    else
        return 0;
}

void
GShipTransporter::changeCargoOnObject(GCargoType type, int32 amount)
{
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        transfer.ore[type] += amount;
        return;
     case el_Supplies:
        transfer.supplies += amount;
        return;
     case el_Colonists:
        transfer.colonists += amount;
        return;
     default:;
    }
}

void
GShipTransporter::changeReservedCargo(GCargoType type, int32 amount)
{
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        reserve.ore[type] += amount;
        return;
     case el_Supplies:
        reserve.supplies += amount;
        return;
     case el_Colonists:
        reserve.colonists += amount;
        return;
     default:;
    }
}

bool
GShipTransporter::canHaveCargo(GCargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Supplies:
     case el_Colonists:
        return true;
     default:
        return false;
    }
}

int32
GShipTransporter::getMaxCargo(GCargoType /*cargo_id*/)
{
    return 10000;               // Tim's turn check
}

string_t
GShipTransporter::getName() const
{
    if (other)
        return other->getName();
    else
        return string_t(_("Jettison"));
}
