/**
  *  \file int/compound.cc
  *  \brief Interpreter: Compound Values
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/compound.h"
#include "int/interror.h"
#include "int/context.h"


bool
IntIndexableValue::isProcedureCall()
{
    return false;
}

IntValue*
IntIndexableValue::call(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    return get(args);
}


/************************ IntSimpleIndexableValue ************************/

/** Constructor.
    \param get  Function to implement get() (=indexed read access, function call).
                Can be null, in this case read access ('pushind' opcode) will not
                be permitted.
    \param make Function to implement makeFirstContext() (=iteration access).
                Can be null, in this case iteration ('firstindex' opcode) will not
                be permitted. */
IntSimpleIndexableValue::IntSimpleIndexableValue(get_t get, make_t make)
    : get_function(get), make_function(make)
{ }

IntSimpleIndexableValue::~IntSimpleIndexableValue()
{ }

// IntIndexableValue:
IntValue*
IntSimpleIndexableValue::get(IntArgBlock& args)
{
    if (get_function) {
        return get_function(args);
    } else {
        throw IntError::typeError(IntError::tyIndexable);
    }
}

void
IntSimpleIndexableValue::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntSimpleIndexableValue::makeFirstContext()
{
    if (make_function)
        return make_function();
    else
        throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntSimpleIndexableValue::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntSimpleIndexableValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleIndexableValue::clone()
{
    return new IntSimpleIndexableValue(*this);
}

/************************ IntSimpleProcedureValue ************************/

IntSimpleProcedureValue::IntSimpleProcedureValue(call_t call)
    : call_function(call)
{ }

IntSimpleProcedureValue::~IntSimpleProcedureValue()
{ }

// IntCallableValue:
IntValue*
IntSimpleProcedureValue::call(IntExecutionContext& exc, IntArgBlock& args)
{
    call_function(exc, args);
    return 0;
}

bool
IntSimpleProcedureValue::isProcedureCall()
{
    return true;
}

// IntValue:
string_t
IntSimpleProcedureValue::toString(bool /*readable*/)
{
    return "#<procedure>";
}

void
IntSimpleProcedureValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleProcedureValue::clone()
{
    return new IntSimpleProcedureValue(call_function);
}


/************************* IntSimpleContextValue *************************/

/** Constructor.
    \param context Contained context. This object will hand out clones of that context.
                   This object will own the context. */
IntSimpleContextValue::IntSimpleContextValue(IntContext* context)
    : context(context)
{ }

IntSimpleContextValue::~IntSimpleContextValue()
{
    delete context;
}

// IntContextValue:
IntContext*
IntSimpleContextValue::makeContext()
{
    return context->clone();
}

// IntValue:
string_t
IntSimpleContextValue::toString(bool /*readable*/)
{
    return "#<context>";
}

void
IntSimpleContextValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleContextValue::clone()
{
    return createFromNewContext(context->clone());
}

/** Constructor. Implements exception-safe construction of
    IntSimpleContextValue. If memory doesn't suffice to create the
    IntSimpleContextValue, the IntContext is deleted as well. */
IntSimpleContextValue*
IntSimpleContextValue::createFromNewContext(IntContext* context)
{
    try {
        return new IntSimpleContextValue(context);
    }
    catch (...) {
        delete context;
        throw;
    }
}

/*************************** IntSubroutineValue **************************/

IntSubroutineValue::IntSubroutineValue(IntBCORef bco)
    : bco(bco)
{ }

IntSubroutineValue::~IntSubroutineValue()
{ }

IntBCORef
IntSubroutineValue::getBytecodeObject()
{
    return bco;
}

string_t
IntSubroutineValue::toString(bool /*readable*/)
{
    return "#<subr>";
}

void
IntSubroutineValue::store(IntTagNode& /*tag*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntSubroutineValue*
IntSubroutineValue::clone()
{
    return new IntSubroutineValue(bco);
}
