/**
  *  \file int/exec.h
  *  \brief Interpreter: Bytecode Process Execution
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_EXEC_H
#define PCC_V2_INT_EXEC_H

#include <memory>
#include "cpluslib/smartptr.h"
#include "int/bytecode.h"
#include "cpluslib/ptrvector.h"
#include "int/context.h"

class IntError;

/** Stack frame of an executing process. */
struct IntExecutionFrame {
    /** Bytecode object executing in this frame. */
    IntBCORef bco;

    /** Next instruction to execute. */
    IntBytecodeObject::pc_t pc;

    /** Local values (parameters, local variables). */
    IntDataSegment local_values;

    /** Local names (parameters, local variables). */
    IntVariableNames local_names;

    /** Top of context stack when this frame was opened. */
    uint32_t context_sp;

    /** Top of exception stack when this frame was opened. */
    uint32_t exception_sp;

    /** Set if caller wants a result on the stack. That is, when this frame is removed, an additional
        value must be pushed to the value stack. */
    bool want_result;

    IntExecutionFrame(IntBCORef bco);
    ~IntExecutionFrame();
};

/** Exception handler. */
struct IntExceptionHandler {
    uint32_t                frame_sp;
    uint32_t                context_sp;
    uint32_t                value_sp;
    IntBytecodeObject::pc_t pc;
};

/** Execution Context (process). */
class IntExecutionContext {
 public:
    IntExecutionContext();
    ~IntExecutionContext();

    void pushNewFrame(IntExecutionFrame* frame);
    void popFrame();

    void pushExceptionHandler(IntBytecodeObject::pc_t pc);
    void popExceptionHandler();
    void handleException(string_t e, string_t trace);

    void pushNewContext(IntContext* ctx);
    void pushContextsFrom(ptr_vector<IntContext>& ctxs);
    void popContext();

    uint32_t getNumActiveFrames() const
        { return frames.size(); }

    IntValue* getResult()
        { return value_stack.top(); }

    void addTraceTo(IntError& err);

    void run();
    void executeInstruction();

    IntContext* lookup(string_t name, int32_t& index);

    static void setNewGlobalContext(IntContext* p);
    static void setBreakHandler(bool check_break());

    bool handleLoad(string_t name);
    bool setVariable(string_t name, IntValue* value);
    GObject* getCurrentObject();

    bool checkForBreak();

 private:
    void checkStack(uint32_t required);
    void handleInvalidOpcode();
    void handleDim(IntDataSegment& values, IntVariableNames& names, uint16_t index);
    void handleFunctionCall(IntBCORef bco, uint32_t nargs, bool want_result, bool expect_proc);
    void handleEvalStatement(uint32_t nargs);
    void handleEvalExpression();
    void handleAddHook();

    /** Termination flag. */
    bool terminated;

    /** All active stack frames. */
    ptr_vector<IntExecutionFrame> frames;

    /** All active contexts for name lookup. */
    ptr_vector<IntContext> contexts;

    /** All active exception handling contexts. */
    ptr_vector<IntExceptionHandler> exceptions;

    /** Value stack. */
    IntDataSegment value_stack;

    /** Global context template. \see setNewGlobalContext */
    static std::auto_ptr<IntContext> global_context;
};

#endif
